package cn.skcks.docking.gb28181.wvp.service.wvp;

import cn.hutool.core.io.IoUtil;
import cn.hutool.crypto.digest.MD5;
import cn.skcks.docking.gb28181.common.json.JsonException;
import cn.skcks.docking.gb28181.common.json.JsonResponse;
import cn.skcks.docking.gb28181.common.json.JsonUtils;
import cn.skcks.docking.gb28181.wvp.config.WvpProxyConfig;
import cn.skcks.docking.gb28181.wvp.dto.device.DeviceChannel;
import cn.skcks.docking.gb28181.wvp.dto.device.GetDeviceChannelsReq;
import cn.skcks.docking.gb28181.wvp.dto.device.GetDeviceChannelsResp;
import cn.skcks.docking.gb28181.wvp.dto.login.WvpLoginReq;
import cn.skcks.docking.gb28181.wvp.dto.login.WvpLoginResp;
import cn.skcks.docking.gb28181.wvp.orm.mybatis.dynamic.model.WvpProxyDevice;
import cn.skcks.docking.gb28181.wvp.proxy.WvpProxyClient;
import cn.skcks.docking.gb28181.wvp.service.device.DeviceService;
import cn.skcks.docking.gb28181.wvp.service.download.DownloadService;
import com.github.rholder.retry.*;
import jakarta.servlet.AsyncContext;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;

import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.text.MessageFormat;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

@Slf4j
@Service
@RequiredArgsConstructor
public class WvpService {
    private final WvpProxyClient wvpProxyClient;
    private final WvpProxyConfig wvpProxyConfig;
    private final DeviceService deviceService;
    private final DownloadService downloadService;

    /**
     * 默认重试次数
     */
    public final static int DEFAULT_RETRY_TIME = 3;
    /**
     * 默认每次重试等待时间
     */
    public final static int DEFAULT_RETRY_WAIT_TIME = 3;

    @SuppressWarnings("UnstableApiUsage")
    private static RetryListener defaultRetryListener(){
        return new RetryListener() {
            @Override
            public <V> void onRetry(Attempt<V> attempt) {
                log.info("第 {} 次 执行结束",attempt.getAttemptNumber());
                if(attempt.hasException()){
                    log.info("异常 {}", attempt.getExceptionCause().getMessage());
                }
            }
        };
    }

    @SuppressWarnings("UnstableApiUsage")
    private static <T> Retryer<JsonResponse<T>> getDefaultRetryer() {
        return RetryerBuilder.<JsonResponse<T>>newBuilder()
                // 异常就重试
                .retryIfException()
                .retryIfRuntimeException()
                // 重试间隔
                .withWaitStrategy(WaitStrategies.fixedWait(DEFAULT_RETRY_WAIT_TIME, TimeUnit.SECONDS))
                // 重试次数
                .withStopStrategy(StopStrategies.stopAfterAttempt(DEFAULT_RETRY_TIME))
                .retryIfResult(result -> result != null && (result.getCode() != 0 || result.getCode() != 200))
                .withRetryListener(defaultRetryListener())
                .build();
    }

    @SuppressWarnings("UnstableApiUsage")
    private static Retryer<JsonResponse<?>> getDefaultGenericRetryer() {
        return RetryerBuilder.<JsonResponse<?>>newBuilder()
                // 异常就重试
                .retryIfException()
                .retryIfRuntimeException()
                // 重试间隔
                .withWaitStrategy(WaitStrategies.fixedWait(DEFAULT_RETRY_WAIT_TIME, TimeUnit.SECONDS))
                // 重试次数
                .withStopStrategy(StopStrategies.stopAfterAttempt(DEFAULT_RETRY_TIME))
                .retryIfResult(result -> result != null && (result.getCode() != 0 || result.getCode() != 200))
                .withRetryListener(defaultRetryListener())
                .build();
    }

    @SneakyThrows
    private void writeErrorToResponse(HttpServletResponse response, JsonResponse<?> json) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        response.setCharacterEncoding(StandardCharsets.UTF_8.name());
        response.setContentType(MediaType.APPLICATION_JSON_VALUE);
        IoUtil.writeUtf8(response.getOutputStream(), false, json);
    }

    @SneakyThrows
    public void video(HttpServletRequest request, HttpServletResponse response, String deviceCode, Date startTime, Date endTime) {
        WvpProxyDevice wvpProxyDevice = deviceService.getDeviceByDeviceCode(deviceCode).orElse(null);
        if (wvpProxyDevice == null) {
            writeErrorToResponse(response, JsonResponse.error("设备不存在"));
            return;
        }
        String deviceId = wvpProxyDevice.getGbDeviceId();
        String channelId = wvpProxyDevice.getGbDeviceChannelId();
        log.info("设备编码 (deviceCode=>{}) 查询到的设备信息 国标id(gbDeviceId => {}), 通道(channelId => {})", deviceCode, deviceId, channelId);

        Retryer<JsonResponse<?>> genericRetryer = getDefaultGenericRetryer();
        AsyncContext asyncContext = request.startAsync();
        asyncContext.setTimeout(0);
        asyncContext.start(() -> {
            HttpServletResponse asyncResponse = (HttpServletResponse) asyncContext.getResponse();
            try {
                genericRetryer.call(() -> video(asyncResponse, deviceCode,deviceId,channelId,startTime,endTime));
            } catch (RetryException e) {
                String reason = MessageFormat.format("查询失败, 已重试 {0} 次", e.getNumberOfFailedAttempts());
                log.error(reason);
                writeErrorToResponse(asyncResponse, JsonResponse.error(reason));
            } catch (Exception e) {
                writeErrorToResponse(asyncResponse, JsonResponse.error(e.getMessage()));
            } finally {
                asyncContext.complete();
            }
        });
    }

    @SneakyThrows
    public JsonResponse<?> video(HttpServletResponse response, String deviceCode, String deviceId,String channelId ,Date startTime, Date endTime) {
        String passwdMd5 = MD5.create().digestHex(wvpProxyConfig.getPasswd());
        WvpLoginReq loginReq = WvpLoginReq.builder()
                .username(wvpProxyConfig.getUser())
                .password(passwdMd5)
                .build();
        JsonResponse<WvpLoginResp> login = wvpProxyClient.login(loginReq);
        String accessToken = login.getData().getAccessToken();
        log.info("wvp 登录成功 accessToken => {}", accessToken);

        log.debug("通过 wvp 查询设备 国标id(gbDeviceId => {}) 通道信息", deviceId);
        JsonResponse<GetDeviceChannelsResp> deviceChannels = wvpProxyClient.getDeviceChannels(accessToken, deviceId, GetDeviceChannelsReq.builder().build());
        if (deviceChannels.getData() == null || deviceChannels.getData().getTotal() == 0) {
            writeErrorToResponse(response, JsonResponse.error(MessageFormat.format("未能获取 设备:{0}, 国标id: {1}, 的通道信息", deviceCode, deviceId)));
            return JsonResponse.success(null);
        }
        List<DeviceChannel> list = deviceChannels.getData().getList();
        log.info("通过 wvp 获取到 查询设备 国标id(gbDeviceId => {}), 通道数量 => {}", deviceId, list.size());
        DeviceChannel deviceChannel = list.parallelStream().filter(item -> item.getChannelId().equalsIgnoreCase(channelId)).findFirst().orElse(null);
        if (deviceChannel == null) {
            writeErrorToResponse(response, JsonResponse.error(MessageFormat.format("未查询到 设备:{0}, 国标id: {1}, 通道: {2} 信息", deviceCode, deviceId, channelId)));
            return JsonResponse.success(null);
        }

        downloadService.download(response, "http://192.168.1.241:18979/download/recordTemp/0490d767d94ce20aedce57c862b6bfe9/rtp/59777645.mp4");
        return login;
    }
}
